"""VGG CNN feature extractor."""

import textwrap
from copy import deepcopy
from typing import Optional, Sequence, Union, List

from torch import nn

from ...cfg import CFG
from ...models._nets import Conv_Bn_Activation
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_maxpool_output_shape,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "VGG16",
]


class VGGBlock(nn.Sequential, SizeMixin):
    """Building blocks of the CNN feature extractor :class:`VGG16`.

    Parameters
    ----------
    num_convs : int
        Number of convolutional layers.
    in_channels : int
        Number of channels in the input.
    out_channels : int
        Number of channels produced by the convolutional layers.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    config : dict
        Other parameters, including
        filter length (kernel size), activation choices,
        weight initializer, batch normalization choices, etc.
        for the convolutional layers;
        and pool size for the pooling layers.

    """

    __name__ = "VGGBlock"

    def __init__(
        self,
        num_convs: int,
        in_channels: int,
        out_channels: int,
        groups: int = 1,
        **config,
    ) -> None:
        super().__init__()
        self.__num_convs = num_convs
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__groups = groups
        self.config = CFG(deepcopy(config))

        self.add_module(
            "cba_1",
            Conv_Bn_Activation(
                in_channels,
                out_channels,
                kernel_size=self.config.filter_length,
                stride=self.config.subsample_length,
                groups=self.__groups,
                activation=self.config.activation,
                kw_activation=self.config.kw_activation,
                kernel_initializer=self.config.kernel_initializer,
                kw_initializer=self.config.kw_initializer,
                norm=self.config.get("norm", self.config.get("batch_norm")),
            ),
        )
        for idx in range(num_convs - 1):
            self.add_module(
                f"cba_{idx+2}",
                Conv_Bn_Activation(
                    out_channels,
                    out_channels,
                    kernel_size=self.config.filter_length,
                    stride=self.config.subsample_length,
                    groups=self.__groups,
                    activation=self.config.activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    norm=self.config.get("norm", self.config.get("batch_norm")),
                ),
            )
        self.add_module(
            "max_pool", nn.MaxPool1d(self.config.pool_size, self.config.pool_stride)
        )

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the module.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            Batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the module.

        """
        num_layers = 0
        for module in self:
            if num_layers < self.__num_convs:
                output_shape = module.compute_output_shape(seq_len, batch_size)
                _, _, seq_len = output_shape
            else:
                output_shape = compute_maxpool_output_shape(
                    input_shape=[batch_size, self.__out_channels, seq_len],
                    kernel_size=self.config.pool_size,
                    stride=self.config.pool_stride,
                    channel_last=False,
                )
            num_layers += 1
        return output_shape


class VGG16(nn.Sequential, SizeMixin, CitationMixin):
    """CNN feature extractor of VGG architecture.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input.
    config : dict
        Other hyper-parameters of the Module, including
        number of convolutional layers, number of filters for each layer,
        and more for :class:`VGGBlock`.
        Key word arguments that have to be set:

            - num_convs: sequence of int,
              number of convolutional layers for each :class:`VGGBlock`.
            - num_filters: sequence of int,
              number of filters for each :class:`VGGBlock`.
            - groups: int,
              connection pattern (of channels) of the inputs and outputs.
            - block: dict,
              other parameters that can be set for :class:`VGGBlock`.

        For a full list of configurable parameters, ref. corr. config file.

    """

    __name__ = "VGG16"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        # self.config = deepcopy(ECG_CRNN_CONFIG.cnn.vgg16)
        self.config = CFG(deepcopy(config))

        module_in_channels = in_channels
        for idx, (nc, nf) in enumerate(
            zip(self.config.num_convs, self.config.num_filters)
        ):
            module_name = f"vgg_block_{idx+1}"
            self.add_module(
                name=module_name,
                module=VGGBlock(
                    num_convs=nc,
                    in_channels=module_in_channels,
                    out_channels=nf,
                    groups=self.config.groups,
                    **(self.config.block),
                ),
            )
            module_in_channels = nf

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the module."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(
            set(
                self.config.get("doi", [])
                + ["10.48550/ARXIV.1409.1556", "10.1016/j.inffus.2019.06.024"]
            )
        )



"""Base class for preprocessors."""

from abc import ABC, abstractmethod
from itertools import repeat
from numbers import Real
from typing import List, Optional, Tuple

import numpy as np
from biosppy.signals.tools import filter_signal
from scipy.ndimage import median_filter

# from scipy.signal import medfilt
# https://github.com/scipy/scipy/issues/9680

from ..utils.misc import ReprMixin, add_docstring
from ..utils.utils_signal import butter_bandpass_filter


__all__ = [
    "PreProcessor",
    "preprocess_multi_lead_signal",
    "preprocess_single_lead_signal",
]


class PreProcessor(ReprMixin, ABC):
    """Base class for preprocessors."""

    __name__ = "PreProcessor"

    @abstractmethod
    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """Apply the preprocessor to `sig`.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.
        fs : numbers.Real
            Sampling frequency of the ECG signal.

        """
        raise NotImplementedError

    @add_docstring(apply)
    def __call__(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """alias of :meth:`self.apply`."""
        return self.apply(sig, fs)

    def _check_sig(self, sig: np.ndarray) -> None:
        """Check validity of the signal.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.

        """
        if sig.ndim not in [1, 2, 3]:
            raise ValueError(
                "Invalid input ECG signal. Should be"
                "1d array, which is a single-lead ECG;"
                "or 2d array, which is a multi-lead ECG of `lead_first` format;"
                "or 3d array, which is a tensor of several ECGs, of shape (batch, lead, siglen)."
            )


def preprocess_multi_lead_signal(
    raw_sig: np.ndarray,
    fs: Real,
    sig_fmt: str = "channel_first",
    bl_win: Optional[List[Real]] = None,
    band_fs: Optional[List[Real]] = None,
    filter_type: str = "butter",
    filter_order: Optional[int] = None,
) -> np.ndarray:
    """Perform preprocessing for multi-lead ECG signal (with units in mV).

    preprocessing may include median filter, bandpass filter, and rpeaks detection, etc.
    Also works for single-lead ECG signal (setting ``sig_fmt="channel_first"``).

    Parameters
    ----------
    raw_sig : numpy.ndarray
        The raw ECG signal, with units in mV.
    fs : numbers.Real
        Sampling frequency of `raw_sig`.
    sig_fmt : str, default "channel_first"
        Format of the multi-lead ECG signal,
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first").
    bl_win : List[numbers.Real], optional
        Window (units in second) of baseline removal
        using :meth:`~scipy.ndimage.median_filter`,
        the first is the shorter one, the second the longer one,
        a typical pair is ``[0.2, 0.6]``.
        If is None or empty, baseline removal will not be performed.
    band_fs : List[numbers.Real], optional
        Frequency band of the bandpass filter,
        a typical pair is ``[0.5, 45]``.
        Be careful when detecting paced rhythm.
        If is None or empty, bandpass filtering will not be performed.
    filter_type : {"butter", "fir"}, optional
        Type of the bandpass filter, default "butter".
    filter_order : int, optional
        Order of the bandpass filter.

    Returns
    -------
    filtered_ecg : numpy.ndarray
        The array of the processed ECG signal.
        The format of the signal is kept the same with the original signal,
        i.e. `sig_fmt`.

    """
    raw_sig = np.asarray(raw_sig)
    assert raw_sig.ndim in [2, 3], "multi-lead signal should be 2d or 3d array"
    assert sig_fmt.lower() in [
        "channel_first",
        "lead_first",
        "channel_last",
        "lead_last",
    ], f"multi-lead signal format `{sig_fmt}` not supported"
    if sig_fmt.lower() in ["channel_last", "lead_last"]:
        # might have a batch dimension at the first axis
        filtered_ecg = np.moveaxis(raw_sig, -2, -1)
    else:
        filtered_ecg = raw_sig

    # remove baseline
    if bl_win:
        window1, window2 = list(repeat(1, filtered_ecg.ndim)), list(
            repeat(1, filtered_ecg.ndim)
        )
        window1[-1] = 2 * (int(bl_win[0] * fs) // 2) + 1  # window size must be odd
        window2[-1] = 2 * (int(bl_win[1] * fs) // 2) + 1
        baseline = median_filter(filtered_ecg, size=window1, mode="nearest")
        baseline = median_filter(baseline, size=window2, mode="nearest")
        filtered_ecg = filtered_ecg - baseline

    # filter signal
    if band_fs:
        assert band_fs[0] < band_fs[1], "Invalid frequency band"
        nyq = 0.5 * fs
        if band_fs[0] <= 0 and band_fs[1] < nyq:
            band = "lowpass"
            frequency = band_fs[1]
        elif band_fs[1] >= nyq and band_fs[0] > 0:
            band = "highpass"
            frequency = band_fs[0]
        elif band_fs[0] > 0 and band_fs[1] < nyq:
            band = "bandpass"
            frequency = band_fs
        else:
            raise AssertionError("Invalid frequency band")
        if filter_type.lower() == "fir":
            filtered_ecg = filter_signal(
                signal=filtered_ecg,
                ftype="FIR",
                # ftype="butter",
                band=band,
                order=filter_order or int(0.2 * fs),
                sampling_rate=fs,
                frequency=frequency,
            )["signal"]
        elif filter_type.lower() == "butter":
            filtered_ecg = butter_bandpass_filter(
                data=filtered_ecg,
                lowcut=band_fs[0],
                highcut=band_fs[1],
                fs=fs,
                order=filter_order
                or round(0.01 * fs),  # better be determined by the `buttord`
            )
        else:
            raise ValueError(f"Unsupported filter type `{filter_type}`")

    if sig_fmt.lower() in ["channel_last", "lead_last"]:
        filtered_ecg = filtered_ecg.T

    return filtered_ecg


def preprocess_single_lead_signal(
    raw_sig: np.ndarray,
    fs: Real,
    bl_win: Optional[List[Real]] = None,
    band_fs: Optional[List[Real]] = None,
    filter_type: str = "butter",
    filter_order: Optional[int] = None,
) -> np.ndarray:
    """Perform preprocessing for single lead ECG signal (with units in mV).

    Preprocessing may include median filter, bandpass filter, and rpeaks detection, etc.

    Parameters
    ----------
    raw_sig : numpy.ndarray
        Raw ECG signal, with units in mV.
    fs : numbers.Real
        Sampling frequency of `raw_sig`.
    bl_win : list (of 2 numbers.Real), optional
        Window (units in second) of baseline removal
        using :meth:`~scipy.ndimage.median_filter`,
        the first is the shorter one, the second the longer one,
        a typical pair is ``[0.2, 0.6]``.
        If is None or empty, baseline removal will not be performed.
    band_fs : list of numbers.Real, optional
        Frequency band of the bandpass filter,
        a typical pair is ``[0.5, 45]``.
        Be careful when detecting paced rhythm.
        If is None or empty, bandpass filtering will not be performed.
    filter_type : {"butter", "fir"}, optional
        Type of the bandpass filter, default "butter".
    filter_order : int, optional,
        Order of the bandpass filter.

    Returns
    -------
    filtered_ecg : numpy.ndarray
        The array of the processed ECG signal.

    NOTE
    ----
    Bandpass filter uses FIR filters, an alternative can be Butterworth filter,
    e.g. :meth:`~torch_ecg.utils.butter_bandpass_filter`.

    """
    filtered_ecg = np.asarray(raw_sig)
    assert filtered_ecg.ndim == 1, "single-lead signal should be 1d array"

    # remove baseline
    if bl_win:
        window1 = 2 * (int(bl_win[0] * fs) // 2) + 1  # window size must be odd
        window2 = 2 * (int(bl_win[1] * fs) // 2) + 1
        baseline = median_filter(filtered_ecg, size=window1, mode="nearest")
        baseline = median_filter(baseline, size=window2, mode="nearest")
        filtered_ecg = filtered_ecg - baseline

    # filter signal
    if band_fs:
        assert band_fs[0] < band_fs[1], "Invalid frequency band"
        nyq = 0.5 * fs
        if band_fs[0] <= 0 and band_fs[1] < nyq:
            band = "lowpass"
            frequency = band_fs[1]
        elif band_fs[1] >= nyq and band_fs[0] > 0:
            band = "highpass"
            frequency = band_fs[0]
        elif band_fs[0] > 0 and band_fs[1] < nyq:
            band = "bandpass"
            frequency = band_fs
        else:
            raise AssertionError("Invalid frequency band")
        if filter_type.lower() == "fir":
            filtered_ecg = filter_signal(
                signal=filtered_ecg,
                ftype="FIR",
                # ftype="butter",
                band=band,
                order=int(0.3 * fs),
                sampling_rate=fs,
                frequency=frequency,
            )["signal"]
        elif filter_type.lower() == "butter":
            filtered_ecg = butter_bandpass_filter(
                data=filtered_ecg,
                lowcut=band_fs[0],
                highcut=band_fs[1],
                fs=fs,
                order=filter_order
                or round(0.01 * fs),  # better be determined by the `buttord`
            )
        else:
            raise ValueError(f"Unsupported filter type `{filter_type}`")

    return filtered_ecg



"""
Higher Order ResNet

References
----------
[1] Luo, Z., Sun, Z., Zhou, W., & Kamata, S. I. (2021). Rethinking ResNets: Improved Stacking Strategies With High Order Schemes. arXiv preprint arXiv:2103.15244.

"""

from torch import nn

from ...cfg import CFG  # noqa: F401
from ...models._nets import (  # noqa: F401
    Activations,
    Conv_Bn_Activation,
    DownSample,
    GlobalContextBlock,
    NonLocalBlock,
    SEBlock,
    ZeroPadding,
)
from ...utils import SizeMixin, CitationMixin


__all__ = [
    "MidPointResNet",
    "RK4ResNet",
    "RK8ResNet",
]


class MidPointResNet(nn.Module, SizeMixin, CitationMixin):
    """ """

    def __init__(self, in_channels: int, **config) -> None:
        """ """
        raise NotImplementedError


class RK4ResNet(nn.Module, SizeMixin, CitationMixin):
    """ """

    def __init__(self, in_channels: int, **config) -> None:
        """ """
        raise NotImplementedError


class RK8ResNet(nn.Module, SizeMixin, CitationMixin):
    """ """

    def __init__(self, in_channels: int, **config) -> None:
        """ """
        raise NotImplementedError



"""
"""

from copy import deepcopy
from numbers import Real
from random import shuffle
from typing import Any, List, Optional, Sequence, Tuple

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter
from ..cfg import DEFAULTS

__all__ = [
    "Mixup",
]


class Mixup(Augmenter):
    """Mixup augmentor.

    Mixup is a data augmentation technique originally proposed in [1]_.
    The PDF file of the paper can be found on arXiv [2]_.
    The official implementation is provided in [3]_. This technique was designed
    for image classification tasks, but it is also widely used for ECG tasks.

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of the ECGs to be augmented.
    alpha : numbers.Real, default 0.5
        alpha parameter of the Beta distribution used in Mixup.
    beta : numbers.Real, optional
        beta parameter of the Beta distribution used in Mixup,
        defaults to `alpha`.
    prob : float, default 0.5
        Probability of applying Mixup.
    inplace : bool, default True
        If True, ECG signal tensors will be modified inplace.
    **kwargs : dict, optional
        Additional keyword arguments, not used.

    Examples
    --------
    .. code-block:: python

        mixup = Mixup(alpha=0.3, beta=0.6, prob=0.7)
        sig = torch.randn(32, 12, 5000)
        label = torch.randint(0, 2, (32, 26), dtype=torch.float32)
        sig, label = mixup(sig, label)

    References
    ----------
    .. [1] Zhang, Hongyi, et al. "mixup: Beyond Empirical Risk Minimization."
           International Conference on Learning Representations. 2018.
    .. [2] https://arxiv.org/abs/1710.09412
    .. [3] https://github.com/facebookresearch/mixup-cifar10/blob/master/train.py

    """

    __name__ = "Mixup"

    def __init__(
        self,
        fs: Optional[int] = None,
        alpha: Real = 0.5,
        beta: Optional[Real] = None,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.alpha = alpha
        self.beta = beta or self.alpha
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Tensor,
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward method of the Mixup augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor
            Label tensor of the ECGs.
        extra_tensors: Sequence[torch.Tensor], optional
            Not used, but kept for consistency with other augmenters.
        **kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor,
            The augmented ECGs.
        label : torch.Tensor
            The augmented labels.
        extra_tensors : Sequence[torch.Tensor], optional
            Unchanged extra tensors.

        """
        batch, lead, siglen = sig.shape
        # TODO: make `lam` different for each batch element, using
        # lam = torch.from_numpy(DEFAULTS.RNG.beta(self.alpha, self.beta, batch), dtype=sig.dtype, device=sig.device)
        # of shape (batch,)
        lam = DEFAULTS.RNG.beta(self.alpha, self.beta)
        indices = np.arange(batch, dtype=int)
        ori = self.get_indices(prob=self.prob, pop_size=batch)
        # print(f"ori = {ori}, len(ori) = {len(ori)}")
        perm = deepcopy(ori)
        shuffle(perm)
        indices[ori] = perm
        indices = torch.from_numpy(indices).long()

        if not self.inplace:
            sig = sig.clone()
            label = label.clone()

        sig = lam * sig + (1 - lam) * sig[indices]
        label = lam * label + (1 - lam) * label[indices]
        # TODO: if `lam` is a Tensor of shape (batch,) instead of a scalar,
        # sig = lam.view(batch, 1, 1) * sig + (1 - lam).view(batch, 1, 1) * sig[indices]
        # label = lam.view(batch, 1) * label + (1 - lam).view(batch, 1) * label[indices]

        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "alpha",
            "beta",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
"""

from numbers import Real
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter

__all__ = [
    "RandomFlip",
]


class RandomFlip(Augmenter):
    """Randomly flip the ECGs along the voltage axis.

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of the ECGs to be augmented
    per_channel : bool, default True
        Whether to flip each channel independently.
    prob : float or Sequence[float], default ``[0.4, 0.2]``
        Probability of performing flip,
        the first probality is for the batch dimension,
        the second probability is for the lead dimension.
    inplace : bool, default True
        If True, ECG signal tensors will be modified inplace.
    kwargs : dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        rf = RandomFlip()
        sig = torch.randn(32, 12, 5000)
        sig, _ = rf(sig, None)

    """

    __name__ = "RandomFlip"

    def __init__(
        self,
        fs: Optional[int] = None,
        per_channel: bool = True,
        prob: Union[Sequence[float], float] = [0.4, 0.2],
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.per_channel = per_channel
        self.inplace = inplace
        self.prob = prob
        if isinstance(self.prob, Real):
            self.prob = np.array([self.prob, self.prob])
        else:
            self.prob = np.array(self.prob)
        assert (self.prob >= 0).all() and (
            self.prob <= 1
        ).all(), "Probability must be between 0 and 1"

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward function of the RandomFlip augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            The ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor, optional
            Label tensor of the ECGs.
            Not used, but kept for consistency with other augmenters.
        extra_tensors : Sequence[torch.Tensor], optional
            Not used, but kept for consistency with other augmenters.
        kwargs : dict, optional
            Additional keyword arguments.
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The augmented ECGs.
        label : torch.Tensor
            The label tensor of the augmented ECGs, unchanged.
        extra_tensors : Sequence[torch.Tensor], optional
            Unchanged extra tensors.

        """
        batch, lead, siglen = sig.shape
        if not self.inplace:
            sig = sig.clone()
        if self.prob[0] == 0:
            return (sig, label, *extra_tensors)
        if self.per_channel:
            flip = torch.ones((batch, lead, 1), dtype=sig.dtype, device=sig.device)
            for i in self.get_indices(prob=self.prob[0], pop_size=batch):
                flip[i, self.get_indices(prob=self.prob[1], pop_size=lead), ...] = -1
            sig = sig.mul_(flip)
        else:
            flip = torch.ones((batch, 1, 1), dtype=sig.dtype, device=sig.device)
            flip[self.get_indices(prob=self.prob[0], pop_size=batch), ...] = -1
            sig = sig.mul_(flip)
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "per_channel",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
the modules that follows CNN feature extractor,
mainly RNN, but can also be attention
"""

from ..cfg import CFG

__all__ = [
    "lstm",
    "attention",
]


lstm = CFG()
lstm.bias = True
lstm.dropouts = 0.2
lstm.bidirectional = True
lstm.retseq = True
lstm.hidden_sizes = [12 * 16, 12 * 16]


attention = CFG()
# almost the same with lstm, but the last layer is an attention layer
attention.head_num = 12
attention.bias = True
attention.dropouts = 0.2
attention.bidirectional = True
attention.hidden_sizes = [12 * 24, 12 * 6]



"""
"""

import warnings
from random import sample
from typing import List, Optional, Tuple

import torch
import torch.nn as nn

from ..utils.misc import ReprMixin
from .bandpass import BandPass
from .baseline_remove import BaselineRemove
from .normalize import Normalize
from .resample import Resample


__all__ = [
    "PreprocManager",
]


class PreprocManager(ReprMixin, nn.Module):
    """Manager class for preprocessors.

    Parameters
    ----------
    pps : Tuple[torch.nn.Module], optional
        The sequence of preprocessors to be added to the manager.
    random : bool, default False
        Whether to apply the preprocessors in random order.
    inplace : bool, default True
        Whether to apply the preprocessors in-place.

    Examples
    --------
    .. code-block:: python

        import torch
        from torch_ecg.cfg import CFG
        from torch_ecg.preprocessors import PreprocManager

        config = CFG(
            random=False,
            bandpass={"fs":500},
            normalize={"method": "min-max"},
        )
        ppm = PreprocManager.from_config(config)
        sig = torch.randn(2, 12, 8000)
        sig = ppm(sig)

    """

    __name__ = "PreprocManager"

    def __init__(
        self,
        *pps: Optional[Tuple[nn.Module, ...]],
        random: bool = False,
        inplace: bool = True,
    ) -> None:
        super().__init__()
        self.random = random
        self._preprocessors = list(pps)

    def _add_bandpass(self, **config: dict) -> None:
        """Add a bandpass filter to the manager.

        Parameters
        ----------
        config : dict
            The configuration of the bandpass filter.

        Returns
        -------
        None

        """
        self._preprocessors.append(BandPass(**config))

    def _add_baseline_remove(self, **config: dict) -> None:
        """Add a median filter for baseline removal to the manager.

        Parameters
        ----------
        config : dict
            The configuration of the median filter.

        Returns
        -------
        None

        """
        self._preprocessors.append(BaselineRemove(**config))

    def _add_normalize(self, **config: dict) -> None:
        """Add a normalizer to the manager.

        Parameters
        ----------
        config : dict
            The configuration of the normalizer.

        Returns
        -------
        None

        """
        self._preprocessors.append(Normalize(**config))

    def _add_resample(self, **config: dict) -> None:
        """Add a resampler to the manager.

        Parameters
        ----------
        config : dict
            The configuration of the resampler.

        Returns
        -------
        None

        """
        self._preprocessors.append(Resample(**config))

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """Apply the preprocessors to the signal tensor.

        Parameters
        ----------
        sig : torch.Tensor
            The signal tensor to be preprocessed.

        Returns
        -------
        torch.Tensor
            The preprocessed signal tensor.

        """
        if len(self.preprocessors) == 0:
            # raise ValueError("No preprocessors added to the manager.")
            # allow dummy (empty) preprocessors
            return sig
        ordering = list(range(len(self.preprocessors)))
        if self.random:
            ordering = sample(ordering, len(ordering))
        for idx in ordering:
            sig = self.preprocessors[idx](sig)
        return sig

    @classmethod
    def from_config(cls, config: dict) -> "PreprocManager":
        """Initialize a :class:`PreprocManager` instance from a configuration.

        Parameters
        ----------
        config : dict
            The configuration of the preprocessors,
            better to be an :class:`~collections.OrderedDict`.

        Returns
        -------
        ppm : PreprocManager
            A new instance of :class:`PreprocManager`.

        """
        ppm = cls(
            random=config.get("random", False), inplace=config.get("inplace", True)
        )
        _mapping = {
            "bandpass": ppm._add_bandpass,
            "baseline_remove": ppm._add_baseline_remove,
            "normalize": ppm._add_normalize,
            "resample": ppm._add_resample,
        }
        for pp_name, pp_config in config.items():
            if pp_name in [
                "random",
                "inplace",
            ]:
                continue
            if pp_name in _mapping and isinstance(pp_config, dict):
                _mapping[pp_name](**pp_config)
            else:
                # just ignore the other items
                pass
                # raise ValueError(f"Unknown preprocessor: {pp_name}")
        if ppm.empty:
            warnings.warn(
                "No preprocessors added to the manager. You are using a dummy preprocessor.",
                RuntimeWarning,
            )
        return ppm

    def rearrange(self, new_ordering: List[str]) -> None:
        """Rearrange the preprocessors.

        Parameters
        ----------
        new_ordering : List[str]
            The new ordering of the preprocessors.

        Returns
        -------
        None

        """
        if self.random:
            warnings.warn(
                "The preprocessors are applied in random order, "
                "rearranging the preprocessors will not take effect.",
                RuntimeWarning,
            )
        _mapping = {  # built-in preprocessors
            "Resample": "resample",
            "BandPass": "bandpass",
            "BaselineRemove": "baseline_remove",
            "Normalize": "normalize",
        }
        _mapping.update({v: k for k, v in _mapping.items()})
        for k in new_ordering:
            if k not in _mapping:
                # allow custom preprocessors
                assert k in [
                    item.__class__.__name__ for item in self._preprocessors
                ], f"Unknown preprocessor name: `{k}`"
                _mapping.update({k: k})
        assert len(new_ordering) == len(
            set(new_ordering)
        ), "Duplicate preprocessor names."
        assert len(new_ordering) == len(
            self._preprocessors
        ), "Number of preprocessors mismatch."
        self._preprocessors.sort(
            key=lambda item: new_ordering.index(_mapping[item.__class__.__name__])
        )

    def add_(self, pp: nn.Module, pos: int = -1) -> None:
        """Add a (custom) preprocessor to the manager.

        This method is preferred against directly manipulating
        the internal list of preprocessors via
        ``PreprocManager.preprocessors.append(pp)``.

        Parameters
        ----------
        pp : torch.nn.Module
            The preprocessor to be added.
        pos : int, default -1
            The position to insert the preprocessor.
            Should be >= -1, with -1 being the indicator of the end.

        """
        assert isinstance(pp, nn.Module)
        assert pp.__class__.__name__ not in [
            p.__class__.__name__ for p in self.preprocessors
        ], f"Preprocessor {pp.__class__.__name__} already exists."
        assert (
            isinstance(pos, int) and pos >= -1
        ), f"pos must be an integer >= -1, but got {pos}."
        if pos == -1:
            self._preprocessors.append(pp)
        else:
            self._preprocessors.insert(pos, pp)

    @property
    def preprocessors(self) -> List[nn.Module]:
        return self._preprocessors

    @property
    def empty(self) -> bool:
        return len(self.preprocessors) == 0

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "random",
            "preprocessors",
        ]
